import functools
from dataclasses import dataclass
from enum import Enum
from enum import auto
from typing import Any
from typing import Optional


class Status(Enum):
    SUCCESS = auto()
    ERROR = auto()


@dataclass
class CheckResult:
    """Store a check result."""

    status: Status

    title: Optional[str] = None
    """The title of the check."""

    description: Optional[str] = None
    """What the check does, and why the spec advises it to do."""

    reason: Optional[str] = None
    """Why it failed, or how it succeed."""

    data: Optional[Any] = None
    """Any related data that can help to debug."""


def decorate_result(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        result = func(*args, **kwargs)
        if isinstance(result, CheckResult):
            result.title = func.__name__
            result.description = func.__doc__
        else:
            result[0].title = func.__name__
            result[0].description = func.__doc__
        return result

    return wrapped



from typing import Tuple

from scim2_client import SCIMClient
from scim2_client import SCIMClientError
from scim2_models import Resource
from scim2_models import ServiceProviderConfig

from .utils import CheckResult
from .utils import Status
from .utils import decorate_result


@decorate_result
def check_service_provider_config_endpoint(
    scim: SCIMClient,
) -> Tuple[Resource, CheckResult]:
    """As described in RFC7644 §4 <rfc7644#section-4>`,
    `/ServiceProviderConfig` is a mandatory endpoint, and should only be
    accessible by GET.

    .. todo::

        Check thet POST/PUT/PATCH/DELETE methods on the endpoint
    """

    try:
        response = scim.query(ServiceProviderConfig)
    except SCIMClientError as exc:
        return CheckResult(status=Status.ERROR, reason=str(exc), data=exc.source)

    return CheckResult(status=Status.SUCCESS, data=response)



import argparse
import uuid
from typing import List
from typing import Tuple

from scim2_client import SCIMClient
from scim2_client import SCIMClientError
from scim2_models import Error
from scim2_models import Group
from scim2_models import Resource
from scim2_models import User

from scim2_tester.resource import check_resource_type
from scim2_tester.resource_types import check_resource_types_endpoint
from scim2_tester.schemas import check_schemas_endpoint
from scim2_tester.service_provider_config import check_service_provider_config_endpoint
from scim2_tester.utils import CheckResult
from scim2_tester.utils import Status
from scim2_tester.utils import decorate_result


@decorate_result
def check_random_url(scim: SCIMClient) -> Tuple[Resource, CheckResult]:
    """A request to a random URL should return a 404 Error object."""

    probably_invalid_url = f"/{str(uuid.uuid4())}"
    try:
        response = scim.query(url=probably_invalid_url)

    except SCIMClientError as exc:
        return CheckResult(status=Status.ERROR, reason=str(exc), data=exc.source)

    if not isinstance(response, Error):
        return CheckResult(
            status=Status.ERROR,
            reason=f"{probably_invalid_url} did not return an Error object",
            data=response,
        )

    if response.status != 404:
        return CheckResult(
            status=Status.ERROR,
            reason=f"{probably_invalid_url} did return an object, but the status code is {response.status}",
            data=response,
        )

    return CheckResult(
        status=Status.SUCCESS,
        reason=f"{probably_invalid_url} correctly returned a 404 error",
        data=response,
    )


def check_server(scim: SCIMClient) -> List[CheckResult]:
    """Perform a series of check to a SCIM server.

    It starts by retrieving the standard :class:`~scim2_models.ServiceProviderConfig`,
    :class:`~scim2_models.Schema` and :class:`~scim2_models.ResourceType` endpoints.

    Then for all discovered resources, it perform a series of creation, query, replacement and deletion.
    """

    results = []

    # Get the initial basic objects
    result_spc = check_service_provider_config_endpoint(scim)
    service_provider_config = result_spc.data
    results.append(result_spc)

    result_schemas = check_schemas_endpoint(scim)
    results.append(result_schemas)

    result_resource_types = check_resource_types_endpoint(scim)
    resource_types = result_resource_types.data
    results.append(result_resource_types)

    # Miscelleaneous checks
    result_random = check_random_url(scim)
    results.append(result_random)

    # Resource checks
    if result_resource_types.status == Status.SUCCESS:
        for resource_type in resource_types:
            results.extend(
                check_resource_type(scim, resource_type, service_provider_config)
            )

    return results


if __name__ == "__main__":
    from httpx import Client

    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument("host")
    parser.add_argument("--token", required=False)
    parser.add_argument("--verbose", required=False, action="store_true")
    args = parser.parse_args()

    client = Client(
        base_url=args.host,
        headers={"Authorization": f"Bearer {args.token}"} if args.token else None,
    )
    scim = SCIMClient(client, resource_types=(User, Group))
    results = check_server(scim)
    for result in results:
        print(result.status.name, result.title)
        if result.reason:
            print("  ", result.reason)
            if args.verbose and result.data:
                print("  ", result.data)



from typing import Tuple

from scim2_client import SCIMClient
from scim2_client import SCIMClientError
from scim2_models import Resource
from scim2_models import Schema

from .utils import CheckResult
from .utils import Status
from .utils import decorate_result


@decorate_result
def check_schemas_endpoint(scim: SCIMClient) -> Tuple[Resource, CheckResult]:
    """As described in RFC7644 §4 <rfc7644#section-4>`, `/ResourceTypes` is a
    mandatory endpoint, and should only be accessible by GET.

    .. todo::

        - Check the POST/PUT/PATCH/DELETE methods on the endpoint
        - Check accessing every subschema with /Schemas/urn:ietf:params:scim:schemas:core:2.0:User
        - Check that query parameters are ignored
        - Check that a 403 response is returned if a filter is passed
        - Check that the 'ResourceType', 'ServiceProviderConfig' and 'Schema' schemas are provided.
    """

    try:
        response = scim.query(Schema)

    except SCIMClientError as exc:
        return CheckResult(status=Status.ERROR, reason=str(exc), data=exc.source)

    available = ", ".join([f"'{resource.name}'" for resource in response.resources])
    return CheckResult(
        status=Status.SUCCESS,
        reason=f"Schemas available are: {available}",
        data=response.resources,
    )



import random
import uuid
from enum import Enum
from typing import List
from typing import Tuple
from typing import get_origin

from pydantic import EmailStr
from scim2_client import SCIMClient
from scim2_client import SCIMClientError
from scim2_models import ComplexAttribute
from scim2_models import Group
from scim2_models import Meta
from scim2_models import Reference
from scim2_models import Resource
from scim2_models import ResourceType
from scim2_models import ServiceProviderConfig
from scim2_models import User

from scim2_tester.utils import CheckResult
from scim2_tester.utils import Status
from scim2_tester.utils import decorate_result


def model_from_resource_type(resource_type: ResourceType):
    """TODO:

    - Actually guess the resource type from the schema.

    """
    if resource_type.id == "User":
        return User

    if resource_type.id == "Group":
        return Group


def fill_with_random_values(obj) -> Resource:
    for field_name, field in obj.model_fields.items():
        if field.default:
            continue

        field_type = obj.get_field_root_type(field_name)
        if field_type is Meta:
            value = None

        elif field_type is int:
            value = uuid.uuid4().int

        elif field_type is bool:
            value = random.choice([True, False])

        elif get_origin(field_type) is Reference:
            value = f"https://{str(uuid.uuid4())}.test"

        elif field_type is EmailStr:
            # pydantic won't allow to use the 'test' TLD here
            value = f"{uuid.uuid4()}@{uuid.uuid4()}.com"

        elif issubclass(field_type, Enum):
            value = random.choice(list(field_type))

        elif issubclass(field_type, ComplexAttribute):
            value = field_type()
            fill_with_random_values(value)

        else:
            value = str(uuid.uuid4())

        # TODO: fix this this is UGLY🤮
        if "list" in str(field.annotation).lower():
            setattr(obj, field_name, [value])

        else:
            setattr(obj, field_name, value)
    return obj


@decorate_result
def check_object_creation(
    scim: SCIMClient, obj: Resource
) -> Tuple[Resource, CheckResult]:
    """Perform an object creation.

    TODO:
      - check if the fields of the result object are the same than the
      fields of the request object
    """

    try:
        response = scim.create(obj)

    except SCIMClientError as exc:
        return CheckResult(status=Status.ERROR, reason=str(exc), data=exc.source)

    return CheckResult(
        status=Status.SUCCESS,
        reason=f"Successfull creation of a {obj.__class__.__name__} object with id {response.id}",
        data=response,
    )


@decorate_result
def check_object_query(scim: SCIMClient, obj: Resource) -> Tuple[Resource, CheckResult]:
    """Perform an object query by knowing its id.

    TODO:
      - check if the fields of the result object are the same than the
      fields of the request object
    """

    try:
        response = scim.query(obj.__class__, obj.id)

    except SCIMClientError as exc:
        return CheckResult(status=Status.ERROR, reason=str(exc), data=exc.source)

    return CheckResult(
        status=Status.SUCCESS,
        reason=f"Successfull query of a {obj.__class__.__name__} object with id {response.id}",
        data=response,
    )


@decorate_result
def check_object_query_without_id(
    scim: SCIMClient, obj: Resource
) -> Tuple[Resource, CheckResult]:
    """Perform an object creation.

    TODO:
      - look for the object across several pages
      - check if the fields of the result object are the same than the
      fields of the request object
    """

    try:
        response = scim.query(obj.__class__)

    except SCIMClientError as exc:
        return CheckResult(status=Status.ERROR, reason=str(exc), data=exc.source)

    found = any(obj.id == resource.id for resource in response.resources)
    if not found:
        return CheckResult(
            status=Status.ERROR,
            reason=f"Could not find object {obj.__class__.__name__} with id : {response.detail}",
            data=response,
        )

    return CheckResult(
        status=Status.SUCCESS,
        reason=f"Successfull query of a {obj.__class__.__name__} object with id {obj.id}",
        data=response,
    )


@decorate_result
def check_object_replacement(
    scim: SCIMClient, obj: Resource
) -> Tuple[Resource, CheckResult]:
    """Perform an object replacement.

    TODO:
      - check if the fields of the result object are the same than the
      fields of the request object
    """

    try:
        response = scim.replace(obj)

    except SCIMClientError as exc:
        return CheckResult(status=Status.ERROR, reason=str(exc), data=exc.source)

    return CheckResult(
        status=Status.SUCCESS,
        reason=f"Successfull replacement of a {obj.__class__.__name__} object with id {response.id}",
        data=response,
    )


@decorate_result
def check_object_deletion(
    scim: SCIMClient, obj: Resource
) -> Tuple[Resource, CheckResult]:
    """Perform an object deletion."""

    try:
        scim.delete(obj.__class__, obj.id)

    except SCIMClientError as exc:
        return CheckResult(status=Status.ERROR, reason=str(exc), data=exc.source)

    return CheckResult(
        status=Status.SUCCESS,
        reason=f"Successfull deletion of a {obj.__class__.__name__} object with id {obj.id}",
    )


def check_resource_type(
    scim: SCIMClient,
    resource_type: ResourceType,
    service_provider_config: ServiceProviderConfig,
) -> List[CheckResult]:
    """
    TODO:

    - Make a 'skip' status
    - Check search
    """
    results = []

    model = model_from_resource_type(resource_type)
    obj = model()
    fill_with_random_values(obj)

    result = check_object_creation(scim, obj)
    results.append(result)

    if result.status == Status.SUCCESS:
        created_obj = result.data
        result = check_object_query(scim, created_obj)
        queried_obj = result.data
        results.append(result)

        result = check_object_query_without_id(scim, created_obj)
        results.append(result)

        fill_with_random_values(queried_obj)
        result = check_object_replacement(scim, created_obj)
        results.append(result)

        result = check_object_deletion(scim, created_obj)
        results.append(result)

    return results



from typing import Tuple

from scim2_client import SCIMClient
from scim2_client import SCIMClientError
from scim2_models import Resource
from scim2_models import ResourceType

from .utils import CheckResult
from .utils import Status
from .utils import decorate_result


@decorate_result
def check_resource_types_endpoint(scim: SCIMClient) -> Tuple[Resource, CheckResult]:
    """As described in RFC7644 §4 <rfc7644#section-4>`, `/ResourceTypes` is a
    mandatory endpoint, and should only be accessible by GET.

    .. todo::

        - Check POST/PUT/PATCH/DELETE on the endpoint
        - Check that query parameters are ignored
        - Check that query parameters are ignored
        - Check that a 403 response is returned if a filter is passed
        - Check that the `schema` attribute exists and is available.
    """

    try:
        response = scim.query(ResourceType)

    except SCIMClientError as exc:
        return CheckResult(status=Status.ERROR, reason=str(exc), data=exc.source)

    available = ", ".join([f"'{resource.name}'" for resource in response.resources])
    reason = f"Resource types available are: {available}"
    return CheckResult(status=Status.SUCCESS, reason=reason, data=response.resources)
